/*
 *  * Copyright 2005-2010 NVIDIA Corporation.  All rights reserved.
 *  */
#ifndef opt_live_INCLUDED
#define opt_live_INCLUDED "opt_live.h"

#include "opt_bb.h"           //BB_LIST_CONTAINER
#include "cxx_memory.h"       //CXX_NEW* && MEM_POOL
#include "opt_main.h"
#include "opt_cfg.h"          //CFG
#include "opt_htable.h"       //CODEREP && STMTREP
#include "map"          //map
#include "bb_node_set.h"
#include "idx_32_set.h"
#include "opt_leaf_iter.h"
#include "opt_mu_chi.h"

using namespace std;

// Class for collecting reference counts of variables and expressions
// in a procedure.
//
class SSA_Refcount
{
  std::map<CODEREP *, INT32> _refcount_map;
  std::set<CODEREP *> _nonRegSet;
  INT32 _var_count;
  void BB_init_cr_refcount(BB_NODE *);
  void CR_mark_non_reg(CODEREP *cr) {
    _nonRegSet.insert(cr);
  }
  void Stmt_init_refcount(STMTREP *stmt);
 
public:
  SSA_Refcount(COMP_UNIT *);

  void CR_collect_refcount(CODEREP *);

  INT32 CR_refcount(CODEREP *cr) {
    if (_refcount_map.find(cr) != _refcount_map.end()) {
      return _refcount_map[cr];
    }
    return 0;
  }

  void Stmt_Dec_refcount(STMTREP *stmt);
  void Stmt_Inc_refcount(STMTREP *stmt);

  void CR_Inc_refcount(CODEREP *cr) {
    if (_refcount_map.find(cr) != _refcount_map.end()) {
      _refcount_map[cr] = _refcount_map[cr] + 1;
    } else {
      _refcount_map[cr] = 1;
    }
    if (cr->Kind() == CK_VAR && _refcount_map[cr] == 2)
      _var_count++;
  }

  void CR_Dec_refcount(CODEREP *cr) {
    if (_refcount_map.find(cr) != _refcount_map.end()) {
      _refcount_map[cr] = _refcount_map[cr] - 1;
    } else {
      _refcount_map[cr] = -1;
    }
  }


  void Expr_Dec_refcount(CODEREP *cr) {
    if (cr->Kind() == CK_VAR) {
      CR_Dec_refcount(cr);
    } else if (cr->Kind() == CK_OP) {
      CR_Dec_refcount(cr);
      for (int i = 0; i < cr->Kid_count(); i++) {
        Expr_Dec_refcount(cr->Opnd(i));
      }
    } else if (cr->Kind() == CK_IVAR) {

      CR_Dec_refcount(cr);
      if (cr->Opr() == OPR_PARM && (cr->Offset() & WN_PARM_DUMMY))
        return;

      if (cr->Ilod_base()) {
        Expr_Dec_refcount(cr->Ilod_base());
      } else if (cr->Istr_base()) {
        Expr_Dec_refcount(cr->Istr_base());
      }

      if (cr->Opr() == OPR_MLOAD) {
        if (cr->Mload_size()) {
          Expr_Dec_refcount(cr->Mload_size());
        } else if (cr->Mload_size()) {
          Expr_Dec_refcount(cr->Mstore_size());
        }
      }
      else if (cr->Opr() == OPR_ILOADX) {
        Expr_Dec_refcount(cr->Index());
      }
    }
  }

  void CR_Inc_refcount_if_exist(CODEREP *cr) {
    if (_refcount_map.find(cr) != _refcount_map.end()) {
      _refcount_map[cr] = _refcount_map[cr] + 1;
      if (cr->Kind() == CK_VAR && _refcount_map[cr] == 2)
        _var_count++;
    }
  }

  INT32 Get_var_count() const
  {
    return _var_count;
  }

  INT32 Generate_Var_Map(std::map<CODEREP *, INT32> &crRegMap, std::map<INT32, CODEREP *> &regCrMap) 
  {
    INT32 count = 0;
    for (std::map<CODEREP *, INT32>::iterator it = _refcount_map.begin(), ie =  _refcount_map.end(); 
      it != ie; ++it) 
    {
      CODEREP *cr = it->first;
      if (_nonRegSet.find(cr) == _nonRegSet.end() && cr->Kind() == CK_VAR && cr->Is_var_volatile() == FALSE) {
        crRegMap[cr] = count;
        regCrMap[count] = cr; 
        count++;
      }
    }

    return count;
  } 
};

// Compute and keep live information for a procedure.
class SSA_Live 

{
  // for a BB_NODE's live information
  typedef struct bb_live_info
  {
    IDX_32_SET *live_in;
    IDX_32_SET *live_out;
    IDX_32_SET *live_gen; 
    IDX_32_SET *live_kill;
    INT32 _live_in_count;

    bb_live_info(INT32 liveSetSize, MEM_POOL *pool) 
    {
      live_in = CXX_NEW(IDX_32_SET(liveSetSize, pool, OPTS_FALSE), pool);
      live_out = CXX_NEW(IDX_32_SET(liveSetSize, pool, OPTS_FALSE), pool);
      live_gen = CXX_NEW(IDX_32_SET(liveSetSize, pool, OPTS_FALSE), pool);
      live_kill = CXX_NEW(IDX_32_SET(liveSetSize, pool, OPTS_FALSE), pool);
    }

  } BB_LIVE_INFO;

  COMP_UNIT *_comp_unit;
  SSA_Refcount *_cr_refcount;
  MEM_POOL _mem_pool;

  INT32 _max_live_in_count;
  std::map<CODEREP *, INT32> _Cr2LiveMap;
  std::map<INT32, CODEREP *> _regCrMap;
 
  INT32 _liveSetSize;      // the size of the set in bb_live_info  
  std::map<BB_NODE *, BB_LIVE_INFO *> _BBLiveMap;  // for each BB_NODE, map to a BB_LIVE_INFO
 
  void Mark_killed(CODEREP *cr, BB_LIVE_INFO *bb_live_info);
  void Mark_up_exposed(CODEREP *cr, BB_LIVE_INFO *bb_live_info);

  INT32 getCRLiveNumber(CODEREP *cr);

  void init_bb_live_info(BB_NODE *bb);
  void Collect_use_rec(CODEREP *cr, BOOL is_store, BB_LIVE_INFO *live_info);
  // build live in for the routine
  void Build_live_in(void);

  INT32 _topVarNumber;   // a temp to keep number of variables for computing live info

  INT32 Compute_bb_live_in(BB_NODE *);

  INT32 Compute_max_live_in();
  INT32 Compute_bb_max_live_count(BB_NODE *bb);

  BOOL CR_is_live_out(CODEREP *cr, BB_LIVE_INFO *blive);
  BOOL CR_is_live_in(CODEREP *cr, BB_LIVE_INFO *blive);

public:

  SSA_Live(COMP_UNIT *cunit, SSA_Refcount *cr_ref);
  ~SSA_Live(void);

  INT32 Get_max_live_in(void) const;

  INT32 Get_bb_live_in(BB_NODE *);

};

extern INT32 Compute_SSA_Max_Live(COMP_UNIT *comp_unit, SSA_Refcount *ref);

#endif
