/*
 * Read output of ir_b2a -st, and create binary intermediate file
 * Note that this is highly dependent on output from ir_reader and symtab
 * print routines.
 */

#include <errno.h>		    /* for sys_errlist */
#include <stdio.h>		    /* for stderr */
#include <string.h>
#include <ctype.h>
#include <map>
#include <stack>
#include <sys/stat.h>
#include "defs.h"
#include "mempool.h"
#include "wn.h"			    /* for ir_reader.h */
#include "wn_simp.h"
#include "wn_util.h"
#include "symtab.h"
#include "const.h"
#include "pu_info.h"
#include "irbdata.h"
#include "ir_reader.h"		    /* for IR_reader_init(), etc. */
#include "ir_bwrite.h"		    /* for WN_open_output(), etc. */
#include "ir_bread.h"		    /* for WN_open_input(), etc. */
#include "erglob.h"
#include "errors.h"
#include "err_host.tab"
#include "config.h"
#include "tracing.h"
#include "dwarf_DST_producer.h"
#include "dwarf_DST_dump.h"


static std::stack<WN*> wn_stack;

extern BOOL 
file_exists (char *path)
{
        INT st;
        struct stat sbuf;
        st = stat(path, &sbuf);
        if (st == -1 && (errno == ENOENT || errno == ENOTDIR))
                return FALSE;
        else
                return TRUE;
}

static void usage (void)
{
  fprintf(stderr, "Usage: ir_a2b <Ascii IR> <Binary IR>\n");
  exit(1);
}

static char line[1024]; // should probably allow any length

static void
internal_error (const char *msg1, const char *msg2)
{
    fprintf(stderr, "ERROR: %s %s\n", msg1, msg2);
    exit(1);
}

#define check_prefix(line,str) \
  if (strncmp(line, str, strlen(str)) != 0) \
    internal_error(line, "does not start with " str)

#define NYI(msg) internal_error("NYI:", msg)

static int
get_line (FILE *f, char *line)
{
        char c;
        char *p = line;
        while (1) {
                c = (char) fgetc(f);
                if (c == (char) EOF) return EOF;
                else if (c == '\n') {
                        *p = '\0';
                        return 0;
                } else {
                        *p = c;
                        p++;
                }
        }
}

// WHIRL utilities

struct ltstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) < 0;
  }
};

static std::map<const char*, OPERATOR, ltstr> operator_map;
static std::map<const char*, INT, ltstr> keyword_map;
static std::map<const char*, TYPE_ID, ltstr> mtype_map;

static INT max_mtype_name_length = 0;

static void
create_mtype_map (void)
{
  /* Enter the mtype names */
  for (INT i = MTYPE_FIRST; i <= MTYPE_LAST; i++) {
    max_mtype_name_length = MAX(max_mtype_name_length, strlen(Mtype_Name(i)));
    mtype_map[Mtype_Name(i)] = i;
  }
}

static void
create_operator_map (void)
{
  /* Enter the WHIRL operators */
  for (INT i = OPERATOR_FIRST; i <= OPERATOR_LAST; i++) {
    OPERATOR opr = (OPERATOR) i;
    operator_map[OPERATOR_name(opr)+strlen("OPR_")] = opr;
  }
}

// some extra pseudo keywords 
enum KEYWORDS {
    KW_END_BLOCK = 1,
    KW_BODY,
    KW_NOOP,
    KW_THEN,
    KW_ELSE,
    KW_END_IF,
    KW_INIT,
    KW_COMP,
    KW_INCR,
    KW_END_COMPGOTO,
    KW_END_XGOTO,
    KW_LOC,
    KW_END_LINFO,
    KW_END_SWITCH,
    KEYWORD_LAST
};

static void
create_keyword_map (void)
{
  /* Enter the extended keywords */
  keyword_map["END_BLOCK"] = KW_END_BLOCK;
  keyword_map["BODY"] = KW_BODY;
  keyword_map["NOOP"] = KW_NOOP;
  keyword_map["THEN"] = KW_THEN;
  keyword_map["ELSE"] = KW_ELSE;
  keyword_map["END_IF"] = KW_END_IF;
  keyword_map["INIT"] = KW_INIT;
  keyword_map["COMP"] = KW_COMP;
  keyword_map["INCR"] = KW_INCR;
  keyword_map["END_COMPGOTO"] = KW_END_COMPGOTO;
  keyword_map["END_XGOTO"] = KW_END_XGOTO;
  keyword_map["LOC"] = KW_LOC;
  keyword_map["END_LINFO"] = KW_END_LINFO;
  keyword_map["END_SWITCH"] = KW_END_SWITCH;
}

static TYPE_ID
get_mtype (char *opcode)
{
  char buf[8];
  TYPE_ID m;
  // search for longest names first, so M8I1 found before M
  for (INT len = max_mtype_name_length; len > 0; --len) {
    strncpy(buf,opcode,len);
    buf[len] = '\0';
    m = mtype_map[buf];
    if (m) {
      return m;
    }
  }
  return MTYPE_UNKNOWN;
}

static ST_IDX
get_st_idx (char *p)
{
  // format is "<level,index,name"
  INT level, index;
  if (*p == '<')
    ++p;
  if (strcmp(p, "null-st") == 0) {
    level = index = 0;
  } else {
    char *q = strchr(p, ',');
    *q = '\0';
    level = atoi(p);
    p = q+1;
    q = strchr(p, ',');
    *q = '\0';
    index = atoi(p);
  }
  return make_ST_IDX(index, level);
}

static TY_IDX
get_ty_idx (char *p)
{
  // format is " T<index,name,align"
  TY_IDX ty;
  INT index;
  UINT align;
  if (strncmp(p, " T<", 3) == 0)
    p += 3;
  char *q = strchr(p, ',');
  *q = '\0';
  index = atoi(p);
  p = q+1;
  q = strchr(p, ',');
  p = q+1;
  align = atoi(p);
  ty = make_TY_IDX(index);
  Set_TY_align(ty, align);
  return ty;
}

static void
read_whirl (FILE *f)
{
  // opcode names don't map well cause are generated not stored
  // so instead look for mtypes and operators
  create_mtype_map();
  create_operator_map();
  create_keyword_map();
  SRCPOS srcpos = 0;

  while (get_line(f, line) != EOF) {
    char *p;
    WN *wn;
    INT key;
    if (line[0] == '\0') continue;
    if (strncmp(line, "=====", 5) == 0) {
      // reached end of function, should be body and params in stack
      // so pop pieces of entry and create whole entry node now.
      WN *body = wn_stack.top();
      wn_stack.pop();
      WN *varrefs = wn_stack.top();
      wn_stack.pop();
      WN *pragmas = wn_stack.top();
      wn_stack.pop();
      INT num_formals = wn_stack.size() - 1;
      WN *formals[8]; // WARNING: tmp hardcoded limit
      for (INT i = num_formals-1; i >= 0; --i) {
        formals[i] = wn_stack.top();
        wn_stack.pop();
      }
      WN *fentry = wn_stack.top();
      WN *entry;
      if (WN_operator(fentry) != OPR_FUNC_ENTRY) 
        fprintf(stderr, "not entry?\n");
      wn_stack.pop();
      // we didn't know how many kids for entry initially, 
      // so reallocate here with correct number.
      entry = WN_CreateEntry (num_formals, WN_st_idx(fentry), body, pragmas, varrefs);
      for (INT i = num_formals-1; i >= 0; --i) {
        WN_formal(entry, i) = formals[i];
      }
      WN_Set_Linenum(entry, WN_Get_Linenum(fentry));
      Set_PU_Info_tree_ptr (Current_PU_Info, entry);
      PU_Info_proc_sym (Current_PU_Info) = WN_st_idx(entry);
      return; // reached end of whirl
    }

    p = strtok(line, " ");
    key = keyword_map[p];
    if (key) {
      WN *block;
      switch (key) {
      case KW_LOC:
        p = strtok(NULL, " ");
        SRCPOS_filenum(srcpos) = atoi(p);
        p = strtok(NULL, " ");
        SRCPOS_linenum(srcpos) = atoi(p);
        break;
      case KW_END_BLOCK:
        // want to pop and insert each stmt into the block;
        // but original block is buried in stack.
        // Could try to find that block, but instead create new block for this.
        block = WN_CreateBlock();
        while (WN_operator(wn_stack.top()) != OPR_BLOCK) {
          WN_INSERT_BlockFirst(block, wn_stack.top());
          wn_stack.pop();
        }
        // now at block in stack
        WN_Set_Linenum(block, WN_Get_Linenum(wn_stack.top()));
        wn_stack.pop();
        wn_stack.push(block);
        break;
      }
    }
    else {
      OPCODE opc;
      INT offset;
      INT id;
      ST_IDX st;
      TY_IDX ty, ty2;
      OPERATOR opr = operator_map[p];
      TYPE_ID rtype, desc;
      if (opr) {
        rtype = desc = MTYPE_V;
      }
      else {
        // find mtypes then operator
        rtype = get_mtype(p);
        if (rtype) {
          p += strlen(Mtype_Name(rtype));
        } else {
          rtype = MTYPE_V;
        }
        desc = get_mtype(p);
        if (desc) {
          p += strlen(Mtype_Name(desc));
        } else {
          desc = MTYPE_V;
        }
        opr = (OPERATOR) operator_map[p];
        if (!opr) fprintf(stderr, "opr not found\n");
        if (OPERATOR_is_store(opr)) {
          // stores have desc not rtype, so swap mtypes
          desc = rtype;
          rtype = MTYPE_V;
        }
      }
      opc = OPCODE_make_op(opr,rtype,desc);
      wn = WN_Create(opr,rtype,desc,OPERATOR_nkids(opr));
      if (OPCODE_has_offset(opc)) {
        p = strtok(NULL, " ");
        offset = atoi(p);
        WN_offset(wn) = offset;
        if (opc == OPC_PRAGMA || opc == OPC_XPRAGMA) {
          // special-case: has flags and then id
          p = strtok(NULL, " ");
          id = atoi(p);
          WN_pragma_flags(wn) = offset;
          WN_pragma(wn) = id;
        }
      }
      if (OPCODE_has_sym(opc)) {
        p = strtok(NULL, ">");
        st = get_st_idx(p);
        WN_st_idx(wn) = st;
      }
      if (OPCODE_has_1ty(opc)) {
        p = strtok(NULL, ">");
        ty = get_ty_idx(p);
        WN_set_ty(wn, ty);
      } else if (OPCODE_has_2ty(opc)) {
        p = strtok(NULL, ">");
        ty = get_ty_idx(p);
        WN_set_ty(wn, ty);
        p = strtok(NULL, ">");
        ty2 = get_ty_idx(p);
        WN_set_load_addr_ty(wn, ty2);
      }
      if (OPCODE_has_next_prev(opc)) {
        WN_Set_Linenum(wn, srcpos);
      }
      if (OPCODE_nkids(opc)) {
        switch (OPCODE_nkids(opc)) {
        case 3:
          WN_kid2(wn) = wn_stack.top();
          wn_stack.pop();
          // fallthru
        case 2:
          WN_kid1(wn) = wn_stack.top();
          wn_stack.pop();
          // fallthru
        case 1:
          WN_kid0(wn) = wn_stack.top();
          wn_stack.pop();
        }
      }
      wn_stack.push(wn);
    }
  }
}

// SYMTAB utilities

static void
read_symtab (FILE *f, SYMTAB_IDX symtab)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "=====", 5) == 0)
      return; // reached end of symtab 
    if (*line == '[') {
      // read symbol
      // e.g. [idx]: name <stidx> class of type name (tyid,mtype) 
      ST *st;
      STR_IDX name;
      ST_CLASS symclass;
      INT tyindex;
      INT tyalign;
      INT flags;
      ST_EXPORT symexport;
      ST_SCLASS sclass;
      TY_IDX tyidx = 0;
      PU_IDX pu;
#ifdef TARG_NVISA
      ST_MEMORY mspace = MEMORY_UNKNOWN;
#endif
      char *p = strtok(line, "]");
      INT idx = atoi(p+1);
      p = strtok(NULL, " ");
      p = strtok(NULL, " \t");
      name = Save_Str(p);
      p = strtok(NULL, " ");
      p = strtok(NULL, " ");
      if (strcmp(p, "Variable") == 0)
        symclass = CLASS_VAR;
      else if (strcmp(p, "Pseudo-Register") == 0)
        symclass = CLASS_PREG;
      else if (strcmp(p, "Subprogram") == 0)
        symclass = CLASS_FUNC;
      else if (strcmp(p, "Constant") == 0)
        symclass = CLASS_CONST;
      else
        fprintf(stderr, "unrecognized class\n");
      p = strtok(NULL, " ");
      if (p && strcmp(p, "of") == 0) {
        p = strtok(NULL, " ");
        p = strtok(NULL, " ");
        p = strtok(NULL, ",");
        p+=2;
        tyidx = make_TY_IDX(atoi(p));
      }
      get_line(f, line);
      // e.g. Address: 0(stidx) Alignment: n bytes
      if (symclass == CLASS_FUNC) {
        p = strtok(line, " \t");
        check_prefix(p, "Returning");
        // skip to pu 
        p = strtok(NULL, "[");
        p = strtok(NULL, "]");
        pu = atoi(p);
        get_line(f, line);
      }
      p = strtok(line, " \t");
      check_prefix(p, "Address:");
      p = strtok(NULL, " ");
      // TODO: base
      p = strtok(NULL, " ");
      if (p && strcmp(p, "Alignment:") == 0) {
        p = strtok(NULL, " ");
        Set_TY_align(tyidx, atoi(p));
      }
      get_line(f, line);
      // e.g. Flags: 0xval, export
      p = strtok(line, " \t");
      check_prefix(p, "Flags:");
      p = strtok(NULL, ",");
#ifdef TARG_NVISA
      if (strstr(p, "__global__"))
        mspace = MEMORY_GLOBAL;
      else if (strstr(p, "__shared__"))
        mspace = MEMORY_SHARED;
      else if (strstr(p, "__constant__"))
        mspace = MEMORY_CONSTANT;
      else if (strstr(p, "__local__"))
        mspace = MEMORY_LOCAL;
      else if (strstr(p, "__param__"))
        mspace = MEMORY_PARAM;
      else if (strstr(p, "__texture__"))
        mspace = MEMORY_TEXTURE;
      else if (strstr(p, "__surface__"))
        mspace = MEMORY_SURFACE;
#endif
      {
        /* p may be hexval<space><flagname>; only want hex val */
        char *q = p;
        while (*q != '\0' && !isspace(*q)) ++q; 
        *q = '\0';
      }
      // flags is hex, so need to convert
      flags = (int) strtol(p, NULL, 16);
      p = strtok(NULL, " ");
      if (strcmp(p, "XLOCAL") == 0)
        symexport = EXPORT_LOCAL;
      else if (strcmp(p, "XPREEMPTIBLE") == 0)
        symexport = EXPORT_PREEMPTIBLE;
      get_line(f, line);
      // e.g. Sclass: sclass
      p = strtok(line, " \t");
      check_prefix(p, "Sclass:");
      p = strtok(NULL, " ");
      if (strcmp(p, "TEXT") == 0)
        sclass = SCLASS_TEXT;
      else if (strcmp(p, "FSTATIC") == 0)
        sclass = SCLASS_FSTATIC;
      else if (strcmp(p, "PSTATIC") == 0)
        sclass = SCLASS_PSTATIC;
      else if (strcmp(p, "COMMON") == 0)
        sclass = SCLASS_COMMON;
      else if (strcmp(p, "REG") == 0)
        sclass = SCLASS_REG;
      else if (strcmp(p, "AUTO") == 0)
        sclass = SCLASS_AUTO;
      else if (strcmp(p, "DGLOBAL") == 0)
        sclass = SCLASS_DGLOBAL;
      else if (strcmp(p, "UGLOBAL") == 0)
        sclass = SCLASS_UGLOBAL;
      else if (strcmp(p, "FORMAL") == 0)
        sclass = SCLASS_FORMAL;
      st = New_ST(symtab);
      ST_Init(st, name, symclass, sclass, symexport, tyidx);
      st->flags = flags;
#ifdef TARG_NVISA
      Set_ST_memory_space (st, mspace);
#endif
      if (symclass == CLASS_FUNC) {
        Set_ST_pu(st, pu);
        // WARNING: for now assume only one function
        Scope_tab[CURRENT_SYMTAB].st = st;
      }
    }
  }
}

static void
read_labels (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "=====", 5) == 0)
      return; // reached end
    NYI("labels");
  }
}

static void
read_pseudoregs (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "=====", 5) == 0)
      return; // reached end
    NYI("pseudo-registers");
  }
}

static void
read_initos (FILE *f, SYMTAB_IDX symtab)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "=====", 5) == 0)
      return; // reached end
    NYI("initos");
  }
}

static void
read_st_attrs (FILE *f, SYMTAB_IDX symtab)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "=====", 5) == 0)
      return; // reached end
    NYI("st_attrs");
  }
}

static void
read_blocks (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "=====", 5) == 0)
      return; // reached end
    NYI("blocks");
  }
}

static void
read_tcons (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "=====", 5) == 0)
      return; // reached end
    if (*line == '[') {
      // read tcon
      // for now, ignore tcons
      // NYI("tcons");
    }
  }
}

// look for pattern of form
// name (#<index>) align <num>
static TY_IDX 
read_ty_idx (char *p)
{
  TY_IDX ty;
  p = strtok(NULL, "#");
  p = strtok(NULL, ")");
  ty = make_TY_IDX(atoi(p));
  p = strtok(NULL, " ");
  check_prefix(p, "align");
  p = strtok(NULL, " ,");
  Set_TY_align(ty, atoi(p));
  return ty;
}

static void
read_pus (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "=====", 5) == 0)
      return; // reached end of symtab 
    if (*line == '[') {
      // read pu 
      // e.g. [idx]: name (#ty) align n, flags 0x
      char *p = strtok(line, "]");
      INT idx = atoi(p+1);
      UINT64 flags;
      INT level;
      TY_IDX tyidx;
      INT srclang;
      INT targetidx;
      INT misc;
      PU_IDX pu_idx;
      PU&    pu = New_PU (pu_idx);
      tyidx = read_ty_idx(p);
      p = strtok(NULL, " ");
      check_prefix(p, "flags");
      p = strtok(NULL, ",");
      // flags is hex, so need to convert
      flags = strtoll(p, NULL, 16);
      get_line(f, line);
      // e.g.  lexical level n, LANG 0x, TARGET_INFO 0,
      p = strtok(line, " \t");
      check_prefix(p, "lexical");
      p = strtok(NULL, " ");
      p = strtok(NULL, ",");
      level = atoi(p);
      p = strtok(NULL, " ");
      check_prefix(p, "LANG");
      p = strtok(NULL, ",");
      srclang = (INT)strtol(p, NULL, 16);
      p = strtok(NULL, " ");
      check_prefix(p, "TARGET_INFO");
      p = strtok(NULL, ",");
      targetidx = atoi(p);
      get_line(f, line);
      // e.g.  Misc. Info (misc) 0
      p = strtok(line, " \t");
      check_prefix(p, "Misc.");
      p = strtok(NULL, " ");
      p = strtok(NULL, " ");
      p = strtok(NULL, " ");
      misc = atoi(p);
      PU_Init (pu, tyidx, level);
      Set_PU_target_idx (pu, targetidx);
      Set_PU_misc_info (pu, misc);
      pu.src_lang = srclang;
      pu.flags = flags;
    }
  }
}

static void
read_types (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "=====", 5) == 0)
      return; // reached end of symtab 
    if (*line == '[') {
      // read pu 
      // e.g. [idx]: name : (f: 0x) size n mtype: kind (mtype)
      char *p = strtok(line, "]");
      TY_IDX idx = atoi(p+1);
      STR_IDX name = 0;
      INT flags;
      INT size;
      char *q;
      TYPE_ID mtype = MTYPE_UNKNOWN;
      TY_KIND kind;
      TY_IDX pty;
      INT puflags;
      TYLIST_IDX tylist;
      TY &ty = New_TY (idx);
      p = strtok(NULL, " ");
      // name can have blanks, so go to : and then find end of last word
      p = strtok(NULL, ":");
      for (q = p+strlen(p)-1; isspace(*q); --q) 
        ;
      ++q;
      *q = '\0';
      if (strcmp(p, "(anon)") != 0) // anon means no ty
        name = Save_Str(p);
      p = strtok(NULL, " ");
      check_prefix(p, "(f:");
      p = strtok(NULL, ")");
      flags = (INT)strtol(p, NULL, 16);
      p = strtok(NULL, " ");
      check_prefix(p, "size");
      p = strtok(NULL, " ");
      size = atoi(p);
      p = strtok(NULL, " ");
      q = p+strlen(p)-1; // end should be :
      check_prefix(q, ":");
      *q = '\0';
      if (p != q) {
        mtype = mtype_map[p];
      }
      p = strtok(NULL, " ");
      if (strcmp(p, "SCALAR") == 0)
        kind = KIND_SCALAR; 
      else if (strcmp(p, "VOID") == 0)
        kind = KIND_VOID; 
      else if (strcmp(p, "->") == 0) {
        kind = KIND_POINTER; 
        pty = read_ty_idx(p);
      }
      else if (strcmp(p, "FUNCTION") == 0) {
        TYLIST_IDX tmplist;
        TY_IDX t;
        char c;
        kind = KIND_FUNCTION; 
        p = strtok(NULL, " ");
        p = strtok(NULL, ")");
        puflags = (INT)strtol(p, NULL, 16);
        get_line(f, line);
        p = strtok(line, " \t");
        check_prefix(p, "returns");
        t = read_ty_idx(p);
        Set_TYLIST_type (New_TYLIST (tylist), t);
        // TODO: need to read arbitrary # parameters until === or [] line
        // use getc;ungetc to peek at first char
        c = fgetc(f);
        while (c == '\t') {
          get_line(f, line);
          p = strtok(line, " ");
          check_prefix(p, "parameter");
          t = read_ty_idx(p);
          Set_TYLIST_type (New_TYLIST (tmplist), t);
          c = fgetc(f);
        }
        ungetc(c, f);
        Set_TYLIST_type (New_TYLIST (tmplist), 0); // end list
      }
      else if (strcmp(p, "ARRAY") == 0) {
        kind = KIND_ARRAY; 
        NYI("array ty");
      }
      else if (strcmp(p, "STRUCT") == 0) {
        kind = KIND_STRUCT; 
        NYI("struct ty");
      }
      TY_Init (ty, size, kind, mtype, name);
      if (TY_kind(ty) == KIND_POINTER)
        Set_TY_pointed(ty, pty);
      else if (TY_kind(ty) == KIND_FUNCTION) {
        Set_TY_tylist(ty, tylist);
        ty.Set_pu_flag((TY_PU_FLAGS)puflags);
      }
    }
  }
}

static void
read_dst_includes (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "-----", 5) == 0)
      return; // reached end
    if (*line == '\0')
      continue;
    if (*line == '[') {
      // read dst
      // e.g. [1,0] ordinal(1) path(.)
      char *p;
      DST_IDX dst;
      p = strtok(line, "(");
      p = strtok(NULL, "(");
      p = strtok(NULL, ")");
      dst = DST_mk_include_dir(p);
    }
  }
}

static void
read_dst_files (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "-----", 5) == 0)
      return; // reached end
    if (*line == '\0')
      continue;
    if (*line == '[') {
      // read dst
      // e.g. [2,0] ordinal(1) name(test.i) path(1) size(187) modt(1285628027)
      char *p;
      DST_IDX dst;
      STR_IDX name;
      UINT path;
      UINT64 size;
      UINT64 modt;
      p = strtok(line, "(");
      p = strtok(NULL, "(");
      p = strtok(NULL, ")");
      name = Save_Str(p); // save it so not lost as process line
      p = strtok(NULL, "(");
      p = strtok(NULL, ")");
      path = atoi(p);
      p = strtok(NULL, "(");
      p = strtok(NULL, ")");
      size = atoll(p);
      p = strtok(NULL, "(");
      p = strtok(NULL, ")");
      modt = atoll(p);
      dst = DST_mk_file_name (Index_To_Str(name), path, size, modt);
    }
  }
}

static void
read_dst_info (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "-----", 5) == 0)
      return; // reached end
    if (*line == '\0')
      continue;
    char *p = line;
    while (isspace(*p)) 
      ++p;
    if (*p == '[') {
      // read dst
    }
  }
}

static void
read_dst_blocks (FILE *f)
{
  while (get_line(f, line) != EOF) {
    if (strncmp(line, "-----", 5) == 0)
      return; // reached end
    if (*line == '\0')
      continue;
    if (strncmp(line, "block ", 6) == 0) {
      // read dst
    }
  }
}

static void
ascii_to_binary (FILE *f)
{
  read_whirl(f);

  // look for current_symtab
  if (get_line(f, line) == EOF) {
    fprintf(stderr, "no symbol table\n");
  }
  
  check_prefix(line, "SYMTAB for ");
  // eventually should grab this function name and later set
  // Scope_tab [Current_scope].st = func_st;
  get_line(f, line);
  check_prefix(line, "=====");
  get_line(f, line); // blank
  get_line(f, line);
  check_prefix(line, "Symbols:");
  read_symtab (f, CURRENT_SYMTAB);
  get_line(f, line);
  check_prefix(line, "Labels:");
  read_labels(f);
  get_line(f, line);
  check_prefix(line, "Pseudo-registers:");
  read_pseudoregs(f);
  get_line(f, line);
  check_prefix(line, "INITOs:");
  read_initos(f, CURRENT_SYMTAB);
  get_line(f, line);
  check_prefix(line, "ST_ATTRs:");
  read_st_attrs(f, CURRENT_SYMTAB);
  get_line(f, line); // blank
  get_line(f, line);
  get_line(f, line);
  check_prefix(line, "GLOBAL SYMTAB:");
  get_line(f, line);
  get_line(f, line);
  check_prefix(line, "File Info:");
  get_line(f, line);
  check_prefix(line, "gp_group:");
  get_line(f, line);
  get_line(f, line);
  check_prefix(line, "Symbols:");
  read_symtab (f, GLOBAL_SYMTAB);
  get_line(f, line);
  check_prefix(line, "PUs:");
  read_pus (f);
  get_line(f, line);
  check_prefix(line, "Types:");
  read_types (f);
  get_line(f, line);
  check_prefix(line, "Blocks:");
  read_blocks(f);
  get_line(f, line);
  check_prefix(line, "Tcons:");
  read_tcons(f);
  get_line(f, line);
  check_prefix(line, "INITOs:");
  read_initos(f, GLOBAL_SYMTAB);
  get_line(f, line);
  check_prefix(line, "ST_ATTRs:");
  read_st_attrs(f, GLOBAL_SYMTAB);
  get_line(f, line);
  check_prefix(line, "String table size");
  get_line(f, line);
  check_prefix(line, "====");
  get_line(f, line); // blank
  get_line(f, line); // blank
  get_line(f, line);
  check_prefix(line, "------------ INCLUDE_DIRECTORIES");
  read_dst_includes(f);
  check_prefix(line, "------------<<<<<<< FILES");
  read_dst_files(f);
  check_prefix(line, "------------<<<<<< DST INFO");
  read_dst_info(f);
  check_prefix(line, "------------<<<< BLOCK INFO");
  read_dst_blocks(f);
}

int main (INT argc, char *argv[])
{
    char *bfile, *afile;
    FILE *bf, *af;
    
    if (argc != 3)
      usage();
    afile = argv[1];
    bfile = argv[2];
    if (!file_exists(afile))
      usage();

    MEM_Initialize();
    Initialize_Symbol_Tables(FALSE);
    // this is from Initialize_Symbol_Tables, but don't create predefined syms:
    Initialize_Strtab (0x1000);
    UINT dummy_idx;
    BZERO (&New_PU ((PU_IDX&) dummy_idx), sizeof(PU));
    BZERO (&New_TY ((TY_IDX&) dummy_idx), sizeof(TY));
    BZERO (New_FLD ().Entry(), sizeof(FLD));
    BZERO (&New_TYLIST ((TYLIST_IDX&) dummy_idx), sizeof(TYLIST));
    BZERO (New_ARB ().Entry(), sizeof(ARB));
    BZERO (&New_BLK ((BLK_IDX&) dummy_idx), sizeof(BLK));
    BZERO (&Initv_Table.New_entry ((INITV_IDX&) dummy_idx), sizeof(INITV));
    Init_Constab ();
    New_Scope (GLOBAL_SYMTAB, Malloc_Mem_Pool, TRUE);
    New_Scope (CURRENT_SYMTAB+1, MEM_pu_nz_pool_ptr, TRUE);
    DST_Init (NULL, 0);
    Current_PU_Info = TYPE_MEM_POOL_ALLOC(PU_Info, Malloc_Mem_Pool);
    PU_Info_init(Current_PU_Info);

    af = fopen(afile, "r");
    ascii_to_binary (af);
    fclose(af);

    // now write the .B file
    Set_PU_Info_state(Current_PU_Info, WT_SYMTAB, Subsect_InMem);
    Set_PU_Info_state(Current_PU_Info, WT_TREE, Subsect_InMem);
    Set_PU_Info_state(Current_PU_Info, WT_PROC_SYM, Subsect_InMem);
    Open_Output_Info (bfile);

    for (PU_Info *pu = Current_PU_Info; pu != NULL; pu = PU_Info_next (pu)) {
        Write_PU_Info (pu);
    }
    Write_Global_Info(Current_PU_Info);
    Close_Output_Info ();

    exit (0);
} /* main */


/* dummy defines for linking */

/* Dummy definitions to satisify references from routines that got pulled
 * in by the header files but are never called
 */
void
Signal_Cleanup (INT sig) { }

const char *
Host_Format_Parm (INT kind, MEM_PTR parm)
{
    fprintf (stderr, "Internal: Host_Format_Parm () not implemented\n");
    return "";
}

INT8 Debug_Level = 0;
